#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <limits.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "xmalloc.h"
#include "vm.h"
#include "task.h"
#include "conf.h"
#include "vmconf.h"
#include "bo.h"

static int key_work_dir(struct cmd *cmd, void *opaque);
static int key_libcall_dir(struct cmd *cmd, void *opaque);
static int key_exec_dir(struct cmd *cmd, void *opaque);
static int key_socket(struct cmd *cmd, void *opaque);
static int key_inetaddr(struct cmd *cmd, void *opaque);
static int key_cycle_timeout(struct cmd *cmd, void *opaque);
static int key_priority(struct cmd *cmd, void *opaque);
static int key_byteorder(struct cmd *cmd, void *opaque);
static int key_intr_num(struct cmd *cmd, void *opaque);
static int key_mode(struct cmd *cmd, void *opaque);
static int key_name(struct cmd *cmd, void *opaque);
static int key_file_name(struct cmd *cmd, void *opaque);

static void section_task(void *opaque);

static struct key_option vm_general_keys[] = {
	{ "work_dir",	   CONF_TYPE_STRING,	key_work_dir },
	{ "libcall_dir",   CONF_TYPE_STRING,	key_libcall_dir },
	{ "exec_dir",	   CONF_TYPE_STRING,	key_exec_dir },
	{ "cycle_timeout", CONF_TYPE_INTEGER,	key_cycle_timeout },
	{ "priority",	   CONF_TYPE_INTEGER,	key_priority },
	{ "socket",	   CONF_TYPE_STRING,	key_socket },
	{ "inet_addr",	   CONF_TYPE_STRING,	key_inetaddr },
	{ "byteorder",	   CONF_TYPE_STRING,	key_byteorder },
	{ NULL,		   CONF_TYPE_UNKNOWN,	NULL }
};

static struct key_option vm_task_keys[] = {
	{ "mode",	CONF_TYPE_STRING,	key_mode },
	{ "file_name",	CONF_TYPE_STRING,	key_file_name },
	{ "name",	CONF_TYPE_STRING,	key_name },
	{ "inter_num",	CONF_TYPE_INTEGER,	key_intr_num },
	{ NULL,		CONF_TYPE_UNKNOWN,	NULL }
};

static struct section_option vm_conf_sections[] = {
	{ "general",	vm_general_keys,	NULL },
	{ "task",	vm_task_keys,		section_task }
};

static struct task_conf *vmopt_get_last_task(struct vm_options *opts)
{
	struct task_conf *iter;

	if (opts->tasks == NULL)
		xerror(1, "Can't locate a task");

	for (iter = opts->tasks; iter->next != NULL; iter = iter->next)
		;

	return iter;
}

static void vmopt_add_task(struct vm_options *opts, struct task_conf *task)
{
	struct task_conf *iter;

	if (opts->tasks != NULL) {
		for (iter = opts->tasks; iter->next != NULL; iter = iter->next)
			;
		iter->next = task;
	} else {
		opts->tasks = task;
	}
}

static void section_task(void *opaque)
{
	struct vm_options *opts;
	struct task_conf *task;

	opts = (struct vm_options *)opaque;

	task = xmalloc0(sizeof(*task));
	task->mode = -1;
	task->intrn = -1;
	vmopt_add_task(opts, task);
}

static int key_work_dir(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;

	opts = (struct vm_options *)opaque;
	xstrncpy(opts->work_dir, cmd->value.str, sizeof(opts->work_dir));

	return 0;
}

static int key_libcall_dir(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts = (struct vm_options *)opaque;

	opts = (struct vm_options *)opaque;
	xstrncpy(opts->libcall_dir, cmd->value.str, sizeof(opts->libcall_dir));

	return 0;
}

static int key_exec_dir(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;

	opts = (struct vm_options *)opaque;
	xstrncpy(opts->exec_dir, cmd->value.str, sizeof(opts->exec_dir));

	return 0;
}

static int key_socket(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;

	opts = (struct vm_options *)opaque;

	if (strlen(opts->inet_addr) != 0) {
		fprintf(stderr, "`socket' and `inet_addr' are mutual exclusive.\n");
		return 1;
	}

	xstrncpy(opts->socket, cmd->value.str, sizeof(opts->socket));

	return 0;
}

static int key_inetaddr(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;

	opts = (struct vm_options *)opaque;

	if (strlen(opts->socket) != 0) {
		fprintf(stderr, "`socket' and `inet_addr' are mutual exclusive.\n");
		return 1;
	}

	xstrncpy(opts->inet_addr, cmd->value.str, sizeof(opts->inet_addr));

	return 0;
}

static int key_cycle_timeout(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;

	opts = (struct vm_options *)opaque;

	if (cmd->value.integer < 0) {
		fprintf(stderr, "Cycle timeout has a negative value.\n");
		return 1;
	}

	opts->cycle_timeout = cmd->value.integer;

	return 0;
}

static int key_priority(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;

	opts = (struct vm_options *)opaque;

	if (cmd->value.integer < 0) {
		fprintf(stderr, "Priority has a negative value.\n");
		return 1;
	}

	opts->priority = cmd->value.integer;

	return 0;
}

static int key_byteorder(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;

	opts = (struct vm_options *)opaque;

	if (!strcasecmp("le", cmd->value.str)) {
		opts->byte_order = BO_LE;
	} else if (!strcasecmp("be", cmd->value.str)) {
		opts->byte_order = BO_BE;
	} else if (!strcasecmp("native", cmd->value.str)) {
		opts->byte_order = BO_NATIVE;
	} else {
		fprintf(stderr, "Unknown byte order value %s.\n", cmd->value.str);
		return -1;
	}

	return 0;
}

static int key_mode(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;
	struct task_conf *task;

	opts = (struct vm_options *)opaque;
	task = vmopt_get_last_task(opts);

	if (!strcmp("cycle", cmd->value.str))
		task->mode = TASK_MODE_CYCLE;
	else if (!strcmp("background", cmd->value.str))
		task->mode = TASK_MODE_BACKGROUND;
	else if (!strcmp("interrupt", cmd->value.str))
		task->mode = TASK_MODE_INTERRUPT;
	else {
		fprintf(stderr, "Unknown mode of task.");
		return 1;
	}

	return 0;
}

static int key_intr_num(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;
	struct task_conf *task;
 
	opts = (struct vm_options *)opaque;
	task = vmopt_get_last_task(opts);

	if (cmd->value.integer < 0 || cmd->value.integer > 31) {
		fprintf(stderr, "Interrupt number takes value from 0 to 31.\n");
		return 1;
	}

	task->intrn = cmd->value.integer;

	return 0;
}

static int key_name(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;
	struct task_conf *task;

	opts = (struct vm_options *)opaque;
	task = vmopt_get_last_task(opts);

	xstrncpy(task->task_name, cmd->value.str, sizeof(task->task_name));

	return 0;
}

static int key_file_name(struct cmd *cmd, void *opaque)
{
	struct vm_options *opts;
	struct task_conf *task;

	opts = (struct vm_options *)opaque;
	task = vmopt_get_last_task(opts);

	xstrncpy(task->file_name, cmd->value.str, sizeof(task->file_name));

	return 0;
}

int read_config(const char *config_file, struct vm_options *opts)
{
	struct config_file *cf;
	int ret;
	
	memset(opts, 0, sizeof(struct vm_options));

	opts->cycle_timeout = -1;
	opts->priority = -1;
	opts->byte_order = -1;

	cf = config_create(config_file, vm_conf_sections, opts);
	if (cf == NULL)
		return -1;

	ret = config_parse(cf);
	config_destroy(&cf);

	return ret;
}

